/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.httpd;

import java.io.IOException;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

/**
 * Calls the working methods of registered {@link Connection}s, if
 * the corresponding channels are ready for the desired operations.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 */
class ConnectionSelector extends Thread {

   private static int _no; // instance counter
   private Selector selector;
   private List queue;

   /**
    * Instanciates and starts this ConnectionSelector.
    */
   public ConnectionSelector() throws IOException {
      super("ConnectionSelector " + (_no++));
      selector = Selector.open();
      queue = new ArrayList();
      start();
   }

   /**
    * Queues a connection and calls {@link Selector#wakeup()}, so that a
    * {@link SelectionKey} can be created and registered.
    *
    * @see #registerQueuedConnections()
    */
   public void enqueue(Connection connection) {
      synchronized (queue) {
         queue.add(connection);
      }
      selector.wakeup();
   }

   /**
    * Registered all queued connections.
    *
    * @see Connection#register(Selector)
    */
   private void registerQueuedConnections() throws IOException {
      if (!queue.isEmpty()) {
         synchronized (queue) {
            while (!queue.isEmpty()) {
               Connection connection = (Connection) queue.remove(queue.size() - 1);
               connection.register(selector);
            }
         }
      }
   }

   /**
    * Calls {@link Selector#select()} in an endless loop.
    */
   public void run() {
      while (true) {
         try {
            int i = selector.select();
            registerQueuedConnections();
            if (i > 0) {
               Set set = selector.selectedKeys();
               Iterator connectionIterator = set.iterator();
               while (connectionIterator.hasNext()) {
                  SelectionKey key = (SelectionKey) connectionIterator.next();
                  Connection connection = (Connection) key.attachment();
                  try {
                     if (key.interestOps() == SelectionKey.OP_READ) {
                        connection.readRequest();
                     } else {
                        connection.writeResponse();
                     }
                  } catch (IOException ioe) {
                     connection.close();
                  } catch (Throwable t) {
                     connection.close();
                     t.printStackTrace();
                  }
               }
            }
         } catch (Throwable t) {
            t.printStackTrace();
         }
      }
   }

}

