/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.memviewer;

import java.awt.*;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.text.NumberFormat;
import java.util.Properties;

/**
 * AWTVisualizer.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 */
public class AWTVisualizer extends Frame implements Visualizer {

   private MemStatesRingBuffer memStates;
   private long maxTotal;
   private double scaleFactor;
   private Polygon totalPolygon;
   private Polygon usedPolygon;
   private int currentFrameHeight;
   private NumberFormat numberFormat;

   public AWTVisualizer() {
      super("MemViewer");
      addWindowListener(new WindowAdapter() {
         public void windowClosing(WindowEvent e) {
            hide();
            dispose();
         }
      });
      numberFormat = NumberFormat.getInstance();
      memStates = new MemStatesRingBuffer();
      totalPolygon = new Polygon();
      usedPolygon = new Polygon();
      setSize(450, 200);
      show();
   }

   public void init(Properties properties) {
      // nothing to do
   }

   public void showMemory(MemState memState) {
      memStates.add(memState);
      update(getGraphics());
   }

   public void update(Graphics g) {
      if (memStates.getLast().getTotalMemory() > maxTotal || getHeight() != currentFrameHeight) {
         maxTotal = memStates.getLast().getTotalMemory();
         currentFrameHeight = getHeight();
         scaleFactor = (double) maxTotal / (double) currentFrameHeight;
      }
      computePolygons();
      super.update(g);
   }

   public void paint(Graphics g) {
      super.paint(g);
      g.setColor(Color.blue);
      g.fillPolygon(totalPolygon);
      g.setColor(Color.red);
      g.fillPolygon(usedPolygon);
      g.setColor(Color.black);
      String line = (numberFormat.format(memStates.getLast().getUsedMemory() / 1024)) + "KB / " + (numberFormat.format(memStates.getLast().getTotalMemory() / 1024)) + "KB";
      g.drawString(line, 10, currentFrameHeight - 10);
      setTitle("MemViewer - " + line);
   }

   private void computePolygons() {
      totalPolygon = new Polygon();
      usedPolygon = new Polygon();
      int x = 0;
      totalPolygon.addPoint(x, currentFrameHeight);
      usedPolygon.addPoint(x, currentFrameHeight);
      for (int i = Math.max(memStates.size() - getWidth(), 0); i < memStates.size(); i++, x++) {
         MemState memState = (MemState) memStates.get(i);
         totalPolygon.addPoint(x, scale(memState.getTotalMemory()));
         usedPolygon.addPoint(x, scale(memState.getUsedMemory()));
      }
      totalPolygon.addPoint(x, currentFrameHeight);
      usedPolygon.addPoint(x, currentFrameHeight);
   }

   private int scale(long memory) {
      return currentFrameHeight - (int) ((double) memory / scaleFactor);
   }

   private static class MemStatesRingBuffer {
      private MemState[] buffer = new MemState[2048];
      private int last;
      private int first;

      public void add(MemState memState) {
         buffer[last] = memState;
         last = ++last % buffer.length;
         if (last == first) first = ++first % buffer.length;
      }

      public MemState get(int index) {
         if (index >= size() || index < 0) throw new IndexOutOfBoundsException("Index: " + index + ", Size: " + size());
         return buffer[(first + index) % buffer.length];
      }

      public MemState getLast() {
         return get(size() - 1);
      }

      public int size() {
         if (first <= last) {
            return last - first;
         }
         return last + (buffer.length - first);
      }

   }
}
