/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.classloading;

/**
 * Loads and initializes classes in a separate thread. You may want to disable
 * class garbage collection with the VM command line parameter <code>-Xnoclassgc</code>
 * in order for this to have an effect.
 * <p>
 * This class is especially useful, when classes shall be loaded over a slow network, but
 * aren't packaged in a zip or jar file. However, consider doing just that (using zip or jar files),
 * before you try to enhance performance by using this class.
 * <p>
 * You may also use this class to preload and preinitialize classes, if you need a lot of classes
 * later on in the life of your application. When you actually have to instantiate a class,
 * you don't need to load and initialize it anymore. This may make applications slightly more
 * responsive.
 * <p>
 * You may obtain a list of classes you might want to preload by running your application with the
 * <code>-verbose:class</code> VM command line parameter.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 * @see java.lang.Class#forName(java.lang.String)
 * @see java.lang.Class#forName(java.lang.String, boolean, java.lang.ClassLoader)
 */
public class BackgroundClassLoader extends Thread {
   private volatile static int count;
   private String[] classnames;
   private boolean initialize;
   private ClassLoader classLoader;

   /**
    * Loads and initializes classes in a separate thread. By default the priority of the thread is
    * {@link Thread#MIN_PRIORITY}. You need to call {@link Thread#start()}, to start loading.
    *
    * @param classnames array of fully qualified classnames
    */
   public BackgroundClassLoader(String[] classnames) {
      this(classnames, true, null);
   }

   /**
    * Loads and optionally initializes classes in a thread. By default the priority of the thread is
    * {@link Thread#MIN_PRIORITY}. You need to call {@link Thread#start()}, to start loading.
    *
    * @param initialize if true, the loaded class is initialized
    * @param classnames array of fully qualified classnames
    * @param classLoader the ClassLoader to use
    */
   public BackgroundClassLoader(String[] classnames, boolean initialize, ClassLoader classLoader) {
      super("BackgroundClassLoader - " + count++);
      if (classnames == null) throw new NullPointerException();
      this.classnames = classnames;
      this.classLoader = classLoader == null ? this.getClass().getClassLoader() : classLoader;
      this.initialize = initialize;
      setDaemon(true);
      setPriority(Thread.MIN_PRIORITY);
   }

   /**
    * Loads the classes. Errors are reported to </code>System.err</code>.
    */
   public void run() {
      for (int i = 0; i < classnames.length; i++) {
         try {
            Class.forName(classnames[i], initialize, classLoader);
         } catch (Exception e) {
            System.err.println(getName() + ": Failed to load class " + classnames[i]);
         }
      }
   }
}
