/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.datastructures;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;

/**
 * Compressed Row Storage Matrix. Suitable for sparse matrices.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 */
public class CRSMatrix extends AbstractMatrix implements Serializable {

   private List entries;
   private int[] rowPointer;

   public CRSMatrix(int cols, int rows) {
      super(cols, rows);
      entries = new ArrayList();
      rowPointer = new int[rows + 1];
      for (int i = 0; i < rows; i++) {
         rowPointer[i] = i;
         // make sure we have the minimum amount of entries in the value list
         entries.add(new Entry(0, null));
      }
      rowPointer[rows] = rows;
   }

   public Object set(int col, int row, Object object) {
      rangeCheck(col, row);
      int pos = getEntryPosition(col, row);
      if (pos >= 0) {
         Entry e = (Entry) entries.get(pos);
         if (e == null) {
            entries.add(pos, new Entry(col, object));
         } else
            return e.setValue(object);
      } else {
         entries.add(-pos - 1, new Entry(col, object));
         for (int i = row + 1; i < rowPointer.length; i++) {
            rowPointer[i]++;
         }
      }
      return null;
   }

   public Object get(int col, int row) {
      rangeCheck(col, row);
      int pos = getEntryPosition(col, row);
      if (pos < 0) return null;
      return getValue((Entry) entries.get(pos));
   }

   public Object remove(int col, int row) {
      rangeCheck(col, row);
      int pos = getEntryPosition(col, row);
      if (pos >= 0) {
         Entry oldEntry = null;
         if (rowPointer[row] + 1 == rowPointer[row + 1]) {
            oldEntry = (Entry) entries.set(pos, new Entry(0, null));
         } else {
            oldEntry = (Entry) entries.remove(pos);
            if (entries instanceof ArrayList) {
               ((ArrayList) entries).trimToSize();
            }
            for (int i = row + 1; i < rowPointer.length; i++) {
               rowPointer[i]--;
            }
         }
         return getValue(oldEntry);
      }
      return null;
   }

   private Object getValue(Entry e) {
      return e == null ? null : e.getValue();
   }

   // binary Search
   private int getEntryPosition(int col, int row) {
      int low = rowPointer[row];
      if (entries.size() > low) {
         int high = rowPointer[row + 1] - 1;

         while (low <= high) {
            int mid = (low + high) >> 1;
            Entry e = (Entry) entries.get(mid);
            int midVal = e.getColIndex();
            int cmp = midVal - col;
            if (cmp < 0)
               low = mid + 1;
            else if (cmp > 0)
               high = mid - 1;
            else
               return mid;
         }
      }
      return -(low + 1);
   }

   private static class Entry {

      int colIndex;
      Object value;

      public Entry(int colIndex, Object value) {
         this.colIndex = colIndex;
         this.value = value;
      }

      public int getColIndex() {
         return colIndex;
      }

      public void setColIndex(int colIndex) {
         this.colIndex = colIndex;
      }

      public Object getValue() {
         return value;
      }

      public Object setValue(Object value) {
         Object oldValue = this.value;
         this.value = value;
         return oldValue;
      }
   }
}
