/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.datastructures;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.ReadableByteChannel;
import java.nio.channels.WritableByteChannel;

/**
 * Same functionality as {@link IntegerSortPipe}, but this implementation uses
 * its own datastructure instead of a {@link java.util.BitSet}.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 * @see IntegerSortPipe
 */
public class IntegerSortPipe1 implements ReadableByteChannel, WritableByteChannel {

   private boolean reading;
   private int offset;
   private ByteBuffer bits;
   private int currentBit;
   private byte currentByte;


   public IntegerSortPipe1(int capacity, int offset) {
      this.offset = offset;
      //bits = ByteBuffer.allocate((int) Math.ceil(((double) capacity) / 8.0d));
      bits = ByteBuffer.wrap(new byte[(int) Math.ceil(((double) capacity) / 8.0d)]);
   }

   public int read(ByteBuffer sortedInts) throws IOException {
      if (!isOpen()) throw new ClosedChannelException();
      if (!reading) {
         bits.rewind();
         reading = true;
      }
      int pos = sortedInts.position();
      // one byte in bits equates to 4*8 bytes in sortedInts
      outerLoop: while ((bits.hasRemaining() || currentBit != 0) && sortedInts.remaining() > 3) {
         if (currentBit == 0) currentByte = bits.get();
         for (; currentBit < 8; currentBit++) {
            if ((currentByte & (1 << currentBit)) != 0) {
               sortedInts.putInt(offset);
            }
            offset++;
            if (sortedInts.remaining() < 4) break outerLoop;
         }
         currentBit = 0;
      }
      int read = sortedInts.position() - pos;
      if (read == 0 && !bits.hasRemaining() && currentBit == 0) {
         reading = false;
         return -1;
      }
      return read;
   }

   public int write(ByteBuffer unsortedInts) throws IOException {
      if (!isOpen() || reading) throw new ClosedChannelException();
      int pos = unsortedInts.position();
      while (unsortedInts.remaining() > 3) {
         int value = unsortedInts.getInt() - offset;
         // only consider the higher three bytes for byteBufferIndex
         int byteBufferIndex = value >> 3;
         byte oldByte = bits.get(byteBufferIndex);
         // set the newByte value
         bits.put(byteBufferIndex, (byte) (oldByte | 1 << (value & 7)));
      }
      return unsortedInts.position() - pos;
   }

   public boolean isOpen() {
      return bits != null;
   }

   public void close() throws IOException {
      if (bits == null) throw new ClosedChannelException();
      bits = null;
   }
}
