/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.datastructures;


/**
 * Cache with random replacement strategy. Which cached item is replaced depends highly
 * on the hashcode of the cached item. The better the hashfunction, the better this cache.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 */
public class RandomCache implements Cache {

   private CacheEntry[] entries;
   private int bitMask;
   private int hits;
   private int misses;

   /**
    * Creates a cache with an initial capacity that will be
    * increased to the next power of 2
    */
   public RandomCache(int initialCapacity) {
      // Find a power of 2 >= initialCapacity
      int capacity = 1;
      while (capacity < initialCapacity)
         capacity <<= 1;
      entries = new CacheEntry[capacity];
      // fill with empty entries
      for (int i = 0; i < capacity; i++) {
         entries[i] = new CacheEntry();
      }
      bitMask = capacity - 1;
   }

   public int getCapacity() {
      return entries.length;
   }

   public Object get(Object key) {
      int index = key.hashCode() & bitMask;
      CacheEntry entry = entries[index];
      if (entry.sameKey(key)) {
         hits++;
         return entry.getValue();
      }
      // no hit
      misses++;
      return null;
   }

   public Object put(Object key, Object value) {
      if (key != null) {
         return entries[key.hashCode() & bitMask].set(key, value);
      }
      return null;
   }

   /**
    * Returns a number between 0.0 and 1.0. 1.0 is a 100% hit rate.
    */
   public float getHitRatio() {
      return ((float) hits) / (float) (hits + misses);
   }

   private static class CacheEntry {
      private Object key;
      private Object value;

      public boolean sameKey(Object other) {
         return key != null && key.equals(other);
      }

      public Object getValue() {
         return value;
      }

      public Object set(Object key, Object value) {
         Object oldValue = value;
         this.key = key;
         this.value = value;
         return oldValue;
      }
   }

}
