/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.datastructures;

import java.util.*;

/**
 * Realization of a SkipList. Note that this is not a complete realization of the {@link Map} interface,
 * some methods aren't supported.<p>
 * A SkipList is an alternative to a balanced tree like {@link TreeMap}. As this implementation is
 * slower than the standard {@link TreeMap} it serves only educational/demonstrational purposes.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 */
public class SkipList extends AbstractMap {

   private static final float DEFAULT_PROBABILITY = 0.25f;
   // probability to increase level
   private float probability;
   // upper bound of levels
   private int maxLevel;
   // greatest level so far
   private int level;
   // the header element of list
   private Entry header;
   // Comparator used.
   private Comparator comparator;
   // size
   private int size;
   // capacity
   private int capacity;
   // cached logarithm
   private double maxLevelDivisor;

   private Entry last;


   /**
    * Constructs a new skip list with a capacity of 101 entries.
    */
   public SkipList() {
      this(101, DEFAULT_PROBABILITY);
   }

   /**
    * Constructs a new skip list optimized for the given capacity.
    */
   public SkipList(int capacity) {
      this(capacity, DEFAULT_PROBABILITY);
   }

   /**
    * Constructs a new skip list optimized for the given capacity and probability.
    */
   public SkipList(int capacity, float probability) {
      this.capacity = capacity;
      this.probability = probability;
      this.maxLevelDivisor = Math.log(1.0 / probability);
      this.maxLevel = getMaxLevel(capacity);
      this.header = new Entry(maxLevel, null, null);
      this.last = header;
   }

   /**
    * Calculates the optimal maximum level for a given capacity.
    */
   private int getMaxLevel(int capacity) {
      return (int) Math.ceil(Math.log(capacity) / maxLevelDivisor);
   }

   /**
    * Generates the level of a new element. The higher a level, the less probable it is.
    */
   private int getRandomLevel() {
      int newLevel = 0;
      while (newLevel < maxLevel && Math.random() < probability) {
         newLevel++;
      }
      return newLevel;
   }


   public Object get(Object key) {
      Entry e = header;
      for (int i = level; i >= 0; i--) {
         for (Entry next = e.next[i]; next != null; next = next.next[i]) {
            int cmp = compare(next.key, key);
            if (cmp < 0)
               e = next;
            else if (cmp > 0)
               break;
            else
               return next.value;
         }
      }
      return null;
   }

   public Object put(Object key, Object value) {
      Entry[] searchPath = getSearchPath(header, key);
      Entry entry = searchPath[0].next[0];
      Object oldValue = null;
      // entry with same key is overwritten
      if (entry != null && key.equals(entry.key)) {
         oldValue = entry.value = value;
      }
      // additional entry is inserted:
      else if (size < capacity) {
         int newLevel = getRandomLevel();
         entry = new Entry(newLevel, key, value);
         // entry has largest level seen in this list: update searchpath
         if (newLevel > level) {
            for (int i = level + 1; i <= newLevel; i++) {
               searchPath[i] = header;
            }
            level = newLevel;
         }
         // insert the entry
         for (int i = 0; i <= newLevel; i++) {
            Entry e = searchPath[i];
            entry.next[i] = e.next[i];
            e.next[i] = entry;
         }
         if (entry.next[0] == null) last = entry;
         size++;
      } else {
         // we need to resize
         resize();
         // call put again, because we need the searchPath for insertion
         oldValue = put(key, value);
      }
      return oldValue;
   }

   public Object remove(Object key) {
      Entry[] searchPath = getSearchPath(header, key);
      Entry entry = searchPath[0].next[0];
      Object oldValue = null;
      if (entry != null && key.equals(entry.key)) {
         oldValue = entry.value;
         for (int i = 0; i <= level; i++) {
            Entry e = searchPath[i];
            if (e.next[i] == entry) {
               e.next[i] = entry.next[i];
            }
            if (entry == last) {
               last = searchPath[0];
            }
         }
         // maybe we have to downsize the level of the list:
         while (header.next[level] == null && level > 0) {
            level--;
         }
         size--;
      }
      return oldValue;
   }

   /**
    * Returns all entries which have keys smaller than the looked for key.
    */
   private Entry[] getSearchPath(Entry e, Object key) {
      Entry[] update = new Entry[maxLevel + 1];
      Entry next;
      for (int i = level; i >= 0; i--) {
         for (next = e.next[i]; next != null && compare(next.key, key) < 0; next = next.next[i]) {
            e = next;
         }
         update[i] = e;
      }
      return update;
   }

   /**
    * Doubles the capacity of this skiplist (if possible) and re-adds all entries.
    */
   private void resize() {
      int newCapacity = capacity << 1;
      if (newCapacity < capacity) newCapacity = Integer.MAX_VALUE;
      capacity = newCapacity;
      maxLevel = getMaxLevel(newCapacity);
      level = 1;
      Entry first = header.next[0];
      header = new Entry(maxLevel, null, null);
      for (Entry e = first; e.next[0] != null; e = e.next[0]) {
         put(e.getKey(), e.getValue());
      }
   }

   /**
    * Compares two keys using the correct comparison method for this SkipList.
    */
   private int compare(Object k1, Object k2) {
      return (comparator == null ? ((Comparable) k1).compareTo(k2)
            : comparator.compare(k1, k2));
   }

   /**
    * Test two values  for equality.  Differs from o1.equals(o2) only in
    * that it copes with with <tt>null</tt> o1 properly.
    */
   private static boolean valEquals(Object o1, Object o2) {
      return (o1 == null ? o2 == null : o1.equals(o2));
   }


   /**
    * Returns the number of key-value mappings in this map.
    *
    * @return the number of key-value mappings in this map.
    */
   public int size() {
      return size;
   }

   /**
    * Returns the first (lowest) key currently in this sorted map.
    *
    * @return the first (lowest) key currently in this sorted map.
    * @throws    NoSuchElementException Map is empty.
    */
   public Object firstKey() {
      Entry first = header.next[0];
      if (first != null) return first.key;
      throw new NoSuchElementException();
   }

   public Object lastKey() {
      if (last != header) return last.key;
      throw new NoSuchElementException();
   }

   /**
    * Returns the comparator used to order this map, or <tt>null</tt> if this
    * map uses its keys' natural order.
    *
    * @return the comparator associated with this sorted map, or
    *                <tt>null</tt> if it uses its keys' natural sort method.
    */
   public Comparator comparator() {
      return comparator;
   }

   public boolean isEmpty() {
      return size == 0;
   }

   public boolean containsKey(Object key) {
      return get(key) != null;
   }

   public void clear() {
      level = 1;
      size = 0;
      header.clear();
   }

   public boolean containsValue(Object value) {
      for (Entry e = header.next[0]; e != null; e = e.next[0]) {
         if (e.value.equals(value)) return true;
      }
      return false;
   }

   /**
    * Not supported.
    */
   public Set entrySet() {
      throw new UnsupportedOperationException();
   }

   /**
    * Not supported.
    */
   public SortedMap subMap(Object fromKey, Object toKey) {
      throw new UnsupportedOperationException();
   }

   /**
    * Not supported.
    */
   public SortedMap headMap(Object toKey) {
      throw new UnsupportedOperationException();
   }

   /**
    * Not supported.
    */
   public SortedMap tailMap(Object fromKey) {
      throw new UnsupportedOperationException();
   }

   private static class Entry implements Map.Entry {

      private Object key;
      private Object value;
      private Entry[] next;

      /**
       * Constructs a new element of a skip list.
       *
       * @param level
       * @param key
       * @param value
       */
      public Entry(int level, Object key, Object value) {
         this.key = key;
         this.value = value;
         this.next = new Entry[level + 1];
      }

      public Object getKey() {
         return key;
      }

      public Object getValue() {
         return value;
      }

      public Object setValue(Object value) {
         Object oldValue = this.value;
         this.value = value;
         return oldValue;
      }

      private void clear() {
         for (int i = 0; i < next.length; i++) {
            next[i] = null;
         }
      }

      public String toString() {
         return key + ":" + value;
      }

   }
}
