/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.io;

import java.io.*;


/**
 * OutputStreamWriter that speeds up ASCII and ISO-8859-1 charsets quite a bit.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 */
public class FastStreamWriter extends OutputStreamWriter {

   private final static String ASCII = "US-ASCII"; // preferred MIME name
   private final static String ISO_8859_1 = "ISO-8859-1"; // preferred MIME name

   private static String ASCII_CANONICAL = ASCII; // canonical name
   private static String ISO_8859_1_CANONICAL = ISO_8859_1; // canonical name

   static {
      try {
         ASCII_CANONICAL = new OutputStreamWriter(new NirvanaOutputStream(), ASCII).getEncoding();
         ISO_8859_1_CANONICAL = new OutputStreamWriter(new NirvanaOutputStream(), ISO_8859_1).getEncoding();
      } catch (UnsupportedEncodingException uee) {
         // can not happen as both encodings have to be
         // supported by every JVM implementation.
         uee.printStackTrace();
      }
   }

   private Writer writer;

   public FastStreamWriter(OutputStream out) throws UnsupportedEncodingException {
      super(out);
      init(out);
   }

   public FastStreamWriter(OutputStream out, String encoding) throws UnsupportedEncodingException {
      super(out, encoding);
      init(out);
   }

   private void init(OutputStream out) throws UnsupportedEncodingException {
      // getEncoding returns the canonical name, therefore we have to compare with
      // the canonical name, too
      if (getEncoding().equals(ASCII_CANONICAL)) {
         writer = new MaskedStreamWriter(out, MaskedStreamWriter.ASCII_MASK);
      } else if (getEncoding().equals(ISO_8859_1_CANONICAL)) {
         writer = new MaskedStreamWriter(out, MaskedStreamWriter.ISO_8859_1_MASK);
      } else {
         writer = new OutputStreamWriter(out, getEncoding());
      }
   }

   public void write(int c) throws IOException {
      writer.write(c);
   }

   public void write(char[] cbuf, int offset, int length) throws IOException {
      writer.write(cbuf, offset, length);
   }

   public void write(String str, int off, int len) throws IOException {
      writer.write(str, off, len);
   }

   public void close() throws IOException {
      writer.close();
   }

   public void flush() throws IOException {
      writer.flush();
   }

}
