/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.io;

import com.tagtraum.perf.datastructures.Cache;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;

/**
 * File cache that uses byte arrays for representing files internally.<p>
 * <em>Note: </em> This implementation is not synchronized.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 */
public class FileCache {

   private Cache cache;
   private long minFilesize;
   private long maxFilesize = Long.MAX_VALUE;
   private int timeToLive = -1;

   /**
    * Initializes this FileCache with a given {@link Cache}-Object.
    */
   public FileCache(Cache cache) {
      this.cache = cache;
   }

   /**
    * Gets a {@link FileCache.Entry} for a {@link File} object.
    */
   public Entry get(File file) throws IOException {
      Entry entry = (Entry) cache.get(file);
      if (entry != null && !entry.isStale()) {
         return entry;
      }
      long length = file.length();
      if (timeToLive != 0 && length > minFilesize && length < maxFilesize) {
         entry = new ByteArrayFile(file);
         // MemoryMappedFile is a lot slower with Sun JDK 1.4.0 on Win2k
         //entry = new MemoryMappedFile(file);
         entry.setTimeToLive(timeToLive);
         cache.put(file, entry);
      }
      else {
         entry = new LiveFile(file);
      }
      return entry;
   }

   /**
    * Minimum size of a file to qualify for caching.
    */
   public long getMinFilesize() {
      return minFilesize;
   }

   /**
    * Minimum size of a file to qualify for caching.
    */
   public void setMinFilesize(long minFilesize) {
      this.minFilesize = minFilesize;
   }

   /**
    * Maximum size of a file to qualify for caching.
    */
   public long getMaxFilesize() {
      return maxFilesize;
   }

   /**
    * Maximum size of a file to qualify for caching.
    */
   public void setMaxFilesize(long maxFilesize) {
      this.maxFilesize = maxFilesize;
   }

   /**
    * Returns the default length of the validity interval in ms without checking, whether
    * the cached content of an entry is still valid. Negative values equate to infinity.
    */
   public int getTimeToLive() {
      return timeToLive;
   }

   /**
    * Sets the default length of the validity interval in ms without checking, whether
    * the cached content of an entry is still valid. Negative values equate to infinity.
    */
   public void setTimeToLive(int timeToLive) {
      this.timeToLive = timeToLive;
   }

   /**
    * FileCache-Entry.
    */
   public static interface Entry {

      /**
       * Returns a <em>new</em> InputStream for the cached file.
       */
      public InputStream getNewInputStream() throws IOException;

      /**
       * Returns the file object this entry has cached.
       */
      public File getFile();

      /**
       * Length of this cached file.
       */
      public long length();

      /**
       * Last modification of this cached file.
       */
      public long lastModified();

      /**
       * Sets the length of the validity interval in ms without checking, whether
       * the cached content is still valid. Negative values equate to infinity.
       */
      public void setTimeToLive(int timeToLive);

      /**
       * Returns the length of the validity interval in ms without checking, whether
       * the cached content is still valid. Negative values equate to infinity.
       */
      public int getTimeToLive();

      /**
       * Indicates whether the cached content is still valid.
       * @return True, if the timetolive is up and either length or lastmodification of the
       *    cached content is not the same as that of the file in the filesystem. False, if either
       *    the timetolive is not up yet or it is up, but the cached content is still the same
       *    as the content of the original file.
       */
      public boolean isStale();
   }

}
