/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.io;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;

/**
 * {@link FileCache.Entry} object that uses memory mapped files.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 */
class MemoryMappedFile implements FileCache.Entry {

   private MappedByteBuffer content;
   private File file;
   private long length;
   private long lastModified;
   private int timeToLive;
   private long creationTime;
   private long lastValidityCheck;

   public MemoryMappedFile(File file) throws IOException {
      this.file = file;
      map(file);
      creationTime = System.currentTimeMillis();
      lastValidityCheck = creationTime;
   }

   private void map(File file) throws IOException {
      this.length = file.length();
      this.lastModified = file.lastModified();
      FileInputStream in = new FileInputStream(file);
      try {
         content = in.getChannel().map(FileChannel.MapMode.READ_ONLY, 0, length);
         content.load();
      } finally {
         if (in != null) try {
            in.close();
         } catch (IOException ioe) {
         }
      }
   }

   public InputStream getNewInputStream() throws IOException {
      return new InputStream() {
         private int pos;

         public int read() throws IOException {
            if (pos == content.limit()) return -1;
            return content.get(pos++);
         }

         public int read(byte b[], int off, int len) throws IOException {
            if (pos == content.limit()) return -1;
            // we need this, because there could be more than one instance of this stream
            // but there is only one instance of the buffer
            synchronized (content) {
               content.position(pos);
               int readBytes = Math.min(len, content.remaining());
               content.get(b, off, readBytes);
               pos += readBytes;
               return readBytes;
            }
         }
      };
   }

   public File getFile() {
      return file;
   }

   public long length() {
      return length;
   }

   public long lastModified() {
      return lastModified;
   }

   public void setTimeToLive(int timeToLive) {
      this.timeToLive = timeToLive;
   }

   public int getTimeToLive() {
      return timeToLive;
   }

   public boolean isStale() {
      boolean stale = false;
      long now = System.currentTimeMillis();
      if (timeToLive >= 0 && lastValidityCheck + (long)timeToLive < now) {
         // time is up, but maybe the thing is still not stale
         stale = length != file.length() || lastModified != file.lastModified();
         if (!stale) lastValidityCheck = now;
      }
      return stale;
   }
}
