/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.io;

import java.io.DataOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.FilterOutputStream;

/**
 * Writes bits to a stream.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 * @see XtendedDataInputStream
 */
public class XtendedDataOutputStream extends DataOutputStream {

   private int bitNumber;
   private byte currentBitByte;
   private boolean firsthalf = true;
   private int currentHalfByte;
   private boolean flushed;

   private OutputStream unfilteredOut;

   public XtendedDataOutputStream(OutputStream out) {
      super(new FlushingFilterOutputStream(out));
      ((FlushingFilterOutputStream)this.out).setXtendedDataOutputStream(this);
      this.unfilteredOut = out;
   }

   /**
    * Writes half a byte to the underlying stream.
    */
   public void writeHalfByte(int hb) throws IOException {
      flushed = false;
      // Set or clear.
      if (firsthalf) {
         currentHalfByte = (hb & 0x0f);
      } else {
         currentHalfByte |= (hb & 0x0f) << 4;
         unfilteredOut.write(currentHalfByte);
      }
      firsthalf = !firsthalf;
   }

   /**
    * Writes a bit to the underlying stream.
    */
   public void writeBit(boolean bitval) throws IOException {
      flushed = false;
      // Set or clear.
      if (bitval)
         currentBitByte |= (byte) (1 << bitNumber);
      else
         currentBitByte &= (byte) (~(1 << bitNumber));
      bitNumber++;
      if (bitNumber == 8) {
         bitNumber = 0;
         unfilteredOut.write(currentBitByte);
         currentBitByte = 0;
      }
   }

   /**
    * Flushes the stream. <p>
    * <em>Note:</em> If less than 8 bits have been written,
    * 8 bits are written nevertheless. The not explicitly set bits
    * default to false.<p>
    * <em>Note:</em> If less than 8 bits have been written,
    * 8 bits are written nevertheless. The not explicitly set bits
    * default to false.
    */
   public void flush() throws IOException {
      bitFlush();
      halfByteFlush();
      super.flush();
   }

   private void bitFlush() throws IOException {
      if (!flushed) {
         if (bitNumber != 0) {
            unfilteredOut.write(currentBitByte);
            currentBitByte = 0;
            bitNumber = 0;
            flushed = true;
         }
      }
   }

   private void halfByteFlush() throws IOException {
      if (!firsthalf) {
         unfilteredOut.write(currentHalfByte);
         firsthalf = true;
      }
   }

   // this is to work around the unpleseant fact that most DataOutputStreams methods are final.
   private static class FlushingFilterOutputStream extends FilterOutputStream {
      private XtendedDataOutputStream xout;
      public FlushingFilterOutputStream(OutputStream out) {
         super(out);
      }

      private void setXtendedDataOutputStream(XtendedDataOutputStream xout) {
         this.xout = xout;
      }

      public void write(int b) throws IOException {
         xout.bitFlush();
         if (!xout.firsthalf) {
            xout.writeHalfByte(b & 0x0f);
            xout.writeHalfByte((b & 0xf0)>>4);
         }
         else {
            super.write(b);
         }
      }

      public void write(byte buf[], int off, int len) throws IOException {
         xout.bitFlush();
         if (!xout.firsthalf) {
            for (int i=off; i<len+off; i++) {
               byte b = buf[i];
               xout.writeHalfByte(b & 0x0f);
               xout.writeHalfByte((b & 0xf0)>>4);
            }
         }
         else {
            super.write(buf, off, len);
         }
      }
   }
}
