/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.servlet;

import javax.servlet.*;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpServletResponseWrapper;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.OutputStreamWriter;
import java.util.zip.GZIPOutputStream;

/**
 * Filter that compresses the output of other filters or Servlets with gzip, if the
 * client is able to process it.
 * <p>
 * <em>Note: </em> This class requires Servlet API 2.3. As this filter closes the OutputStream
 * it must be the first in a chain.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 * @see CompressorServlet
 */
public class GZIPFilter implements Filter {
   public void doFilter(ServletRequest request, final ServletResponse response, FilterChain chain) throws IOException, ServletException {
      HttpServletRequest req = (HttpServletRequest) request;
      HttpServletResponse res = (HttpServletResponse) response;
      if (acceptsGzip(req)) {
         res = new HttpServletResponseWrapper(res) {
            private ServletOutputStream wrappedOut;
            private PrintWriter wrappedWriter;

            public ServletOutputStream getOutputStream() throws IOException {
               if (wrappedOut == null) {
                  setHeader("Content-Encoding", "gzip");
                  final OutputStream out = new GZIPOutputStream(response.getOutputStream());
                  wrappedOut = new ServletOutputStream() {
                     public void write(int b) throws IOException {
                        out.write(b);
                     }

                     public void write(byte b[], int off, int len) throws IOException {
                        out.write(b, off, len);
                     }

                     public void close() throws IOException {
                        out.close();
                     }
                  };
               }
               return wrappedOut;
            }

            public PrintWriter getWriter() throws IOException {
               if (wrappedWriter == null) {
                  wrappedWriter = new PrintWriter(new OutputStreamWriter(getOutputStream(), getCharacterEncoding()));
               }
               return wrappedWriter;
            }

         };
         try {
            chain.doFilter(req, res);
         } finally {
            // make sure that the stream will be properly closed, as this is essential for GZIP.
            res.getWriter().flush();
            res.getOutputStream().close();
         }
      }
      else {
         chain.doFilter(req, res);
      }
   }

   private boolean acceptsGzip(HttpServletRequest request) {
      String acceptEncoding = request.getHeader("Accept-Encoding");
      return acceptEncoding != null && acceptEncoding.indexOf("gzip") != -1;
   }

   public void init(FilterConfig config) throws ServletException {
   }

   public void destroy() {
   }
}
