/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.strings;

import java.text.DateFormat;
import java.text.FieldPosition;
import java.text.ParsePosition;
import java.util.Calendar;
import java.util.Date;
import java.util.TimeZone;

/**
 * Parses a {@link Date} in the format "HH:mm:ss,SSS" for example, "15:49:37,459".
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 */
public class AbsoluteTimeDateFormat extends DateFormat {

    private String lastString;
    private int lastOffset;
    private int lastHour;
    private int lastMinute;
    private int lastSecond;


    public AbsoluteTimeDateFormat() {
        this(TimeZone.getDefault());
    }

    public AbsoluteTimeDateFormat(TimeZone timeZone) {
        setCalendar(Calendar.getInstance(timeZone));
        lastString = "";
    }

    public StringBuffer format(Date date, StringBuffer toAppendTo,
                               FieldPosition fieldPosition) {
        // delegate to e.g. log4J version of this method
        return null;
    }

    public Date parse(String s, ParsePosition p) {
        calendar.clear();
        try {
            subParse(s, p);
        } catch (RuntimeException e) {
            p.setErrorIndex(p.getIndex());
            return null;
        }
        return calendar.getTime();
    }

    protected void subParse(String s, ParsePosition p) {
        if (!getCachedTime(s, p)) {
            int offset = p.getIndex();
            parseTime(s, p);
            putCachedTime(s, offset);
        }
    }

    /**
     * Parses the time portion of the given String and sets its values in the current calendar object.
     */
    private void parseTime(String s, ParsePosition p) {
        calendar.set(Calendar.HOUR_OF_DAY, NumberParser.parseInt(s, p.getIndex(), 2));
        calendar.set(Calendar.MINUTE, NumberParser.parseInt(s, p.getIndex() + 3, 2));
        calendar.set(Calendar.SECOND, NumberParser.parseInt(s, p.getIndex() + 6, 2));
        calendar.set(Calendar.MILLISECOND, NumberParser.parseInt(s, p.getIndex() + 9, 3));
        p.setIndex(p.getIndex() + 12);
    }

    /**
     * Returns true if a cached time could be obtained and set in the current calendar object.
     */
    private boolean getCachedTime(String s, ParsePosition p) {
        // relevant is only the part before the milliseconds as that is more likely to be the same
        if (s.regionMatches(p.getIndex(), lastString, lastOffset, 9)) {
            calendar.set(Calendar.HOUR_OF_DAY, lastHour);
            calendar.set(Calendar.MINUTE, lastMinute);
            calendar.set(Calendar.SECOND, lastSecond);
            calendar.set(Calendar.MILLISECOND, NumberParser.parseInt(s, p.getIndex() + 9, 3));
            p.setIndex(p.getIndex() + 12);
            return true;
        }
        return false;
    }

    /**
     * Associates the date portion of the current calendar object with the string.
     */
    private void putCachedTime(String s, int offset) {
        lastString = s;
        lastOffset = offset;
        lastHour = calendar.get(Calendar.HOUR_OF_DAY);
        lastMinute = calendar.get(Calendar.MINUTE);
        lastSecond = calendar.get(Calendar.SECOND);
    }

}
