/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.strings;

import java.io.Serializable;
import java.util.*;

/**
 * Case insensitive map.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 */
public class CaseInsensitiveMap implements Map, Serializable {
    private Map map;
    private Map upperCaseMap;
    private Map equivalentKeys;

    public CaseInsensitiveMap() {
        this.map = new HashMap();
        this.upperCaseMap = new HashMap();
        this.equivalentKeys = new HashMap();
    }

    // string specific methods

    public Object get(String key) {
        Object value = map.get(key);
        return value != null ? value : upperCaseMap.get(key.toUpperCase());
    }

    public Object remove(String key) {
        Object oldValue = null;
        if (key != null) {
            String upperCaseKey = key.toUpperCase();
            oldValue = upperCaseMap.remove(upperCaseKey);
            Set s = (Set)equivalentKeys.get(upperCaseKey);
            if (s != null) {
                for (Iterator i = s.iterator(); i.hasNext(); ) {
                    map.remove(i.next());
                }
            }
        }
        else {
            oldValue = upperCaseMap.remove(null);
            map.remove(null);
        }
        return oldValue;
    }

    public Object put(String key, Object value) {
        Object oldValue = null;
        if (key != null) {
            String upperCaseKey = key.toUpperCase();
            oldValue = upperCaseMap.get(upperCaseKey);
            if (oldValue != value) {
                remove(key);
            }
            upperCaseMap.put(upperCaseKey, value);
            map.put(key, value);
            Set set = (Set)equivalentKeys.get(upperCaseKey);
            if (set == null) {
                set = new HashSet();
                equivalentKeys.put(upperCaseKey, set);
            }
            set.add(key);
        }
        else {
            oldValue = upperCaseMap.put(null, value);
            map.put(null, value);
        }
        return oldValue;
    }

    public boolean containsKey(String key) {
        boolean contains = map.containsKey(key);
        return key != null ? upperCaseMap.containsKey(key.toUpperCase()) : contains;
    }

    // other methods

    public int size() {
        return upperCaseMap.size();
    }

    public boolean isEmpty() {
        return map.isEmpty();
    }

    public boolean containsKey(Object key) {
        return containsKey((String) key);
    }

    public boolean containsValue(Object value) {
        return map.containsValue(value);
    }

    public Object get(Object key) {
        return get((String) key);
    }

    public Object put(Object key, Object value) {
        return put((String) key, value);
    }

    public Object remove(Object key) {
        return remove((String) key);
    }

    public void putAll(Map t) {
        Iterator i = t.entrySet().iterator();
        while (i.hasNext()) {
            Map.Entry e = (Map.Entry) i.next();
            put(e.getKey(), e.getValue());
        }
    }

    public void clear() {
        map.clear();
        upperCaseMap.clear();
    }

    public Set keySet() {
        return upperCaseMap.keySet();
    }

    public Collection values() {
        return upperCaseMap.values();
    }

    public Set entrySet() {
        return upperCaseMap.entrySet();
    }

    public int hashCode() {
        return upperCaseMap.hashCode();
    }

    public boolean equals(Object obj) {
        return upperCaseMap.equals(obj);
    }

    public String toString() {
        return upperCaseMap.toString();
    }

}
