/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.util;

import com.tagtraum.perf.io.XtendedDataOutputStream;
import com.tagtraum.perf.io.XtendedDataInputStream;

import java.io.*;
import java.util.ResourceBundle;

/**
 * Implementation of a Huffmantree. Note that this class is not optimized.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 */
public class HuffmanTree {

   private Node[] nodes;
   private int treesize;
   private static ResourceBundle localStrings = ResourceBundle.getBundle("com.tagtraum.perf.util.localStrings");

   public void encode(byte[] plaintext, int offset, int length, OutputStream rawOut) throws IOException {
      nodes = new Node[512];
      treesize = 0;
      buildLeafs(plaintext, offset, length);
      buildTree();
      DataOutputStream out = new DataOutputStream(rawOut);
      writeTree(out);
      // write compressed stuff
      writeContent(plaintext, offset, length, out);
   }

   private void writeTree(DataOutputStream out) throws IOException {
      // write treesize and implicitly the root node number (i.e. the last node is the root node)
      out.writeShort(treesize);
      // write tree nodes
      for (int i = 0; i < 512; i++) {
         if (nodes[i] != null) nodes[i].write(out);
      }
   }

   private void buildLeafs(byte[] plaintext, int offset, int length) {
      // instantiate leafs and count frequency
      for (int i = offset, limit = offset + length; i < limit; i++) {
         int index = plaintext[i] & 0xff;
         Node node = nodes[index];
         if (node == null) {
            node = new Node(index); // implicitly sets the frequency to 1.0f
            nodes[index] = node;
            treesize++;
         }
         else {
            node.incFrequency();
         }
      }
      // normalize frequencies
      for (int i = 0; i < 256; i++) {
         if (nodes[i] != null) nodes[i].normalizeFrequency((float) length);
      }
   }

   /**
    * @param root number of leafs
    * @return number of nodes in the full tree
    */
   private void buildTree() {
      // Go through the tree. Locate nodes of really low frequency.
      int root = 256;
      while (true) {
         int leftChild = findLowestFrequency(root, -1);
         // Did we find a lowest value? If not, the tree is done.
         if (leftChild == -1) break;
         float leftFrequency = nodes[leftChild].getFrequency();

         int rightChild = findLowestFrequency(root, leftChild);
         // Did we find the next lowest value? If not, the tree is done.
         if (rightChild == -1) break;
         float rightFrequency = nodes[rightChild].getFrequency();

         // Make a new root, and attach the two selected nodes to it.
         // New root.
         nodes[leftChild].setParent(root);
         nodes[rightChild].setParent(root);
         nodes[root] = new Node(leftChild, rightChild, leftFrequency + rightFrequency, root);
         root++;
         treesize++;
      }
   }

   private int findLowestFrequency(int root, int exception) {
      // Find first-lowest frequency.
      float lf = 2.0f;
      int index = -1;
      for (int i = 0; i < root; i++) {
         Node currentNode = nodes[i];
         if (currentNode != null && currentNode.getParent()<0 && (i != exception) && currentNode.getFrequency() <= lf) {
            lf = currentNode.getFrequency();
            index = i;
         }
      }
      return index;
   }


   private void writeContent(byte[] plaintext, int offset, int length, DataOutputStream rawOut) throws IOException {
      // write number of symbols
      rawOut.writeInt(length);
      XtendedDataOutputStream bitOut = new XtendedDataOutputStream(rawOut);
      // Huffman tree is built. Compress the plaintext.
      byte[] bitstring = new byte[30];
      for (int i = offset, max = offset + length; i < max; i++) {
         // Fetch a "character."
         int c = plaintext[i] & 0xff;
         // Build a bit string for byte value.
         int bitstringlen = 0;
         // ...until you hit the root.
         int parent;
         while ((parent = nodes[c].getParent()) != -2) {
            if (nodes[parent].getLeft() == c)
               bitstring[bitstringlen] = 0;
            else
               bitstring[bitstringlen] = 1;
            c = parent;
            bitstringlen++;
         }
         // Step backward through the bit string, writing
         // the bits to the BitStream
         while (0 != bitstringlen--) {
            bitOut.writeBit(bitstring[bitstringlen] == 1);
         }
      }
      bitOut.flush();
   }


   public byte[] decode(InputStream in) throws IOException {
      DataInputStream dataInput = new DataInputStream(in);
      int root = readTree(dataInput);
      return readContent(root, dataInput);
   }

   private int readTree(DataInputStream in) throws IOException {
      // read root node number
      treesize = in.readUnsignedShort();
      int index = 0;
      if (treesize == -1) throw new EOFException();
      try {
         nodes = new Node[512];
         // read tree nodes
         for (int i = 0; i < treesize; i++) {
            Node node = new Node(in);
            index = node.getValue();
            nodes[index] = node;
         }
      }
      catch (EOFException eof) {
         throw new HuffmanException(localStrings.getString("unexpected_eof"));
      }
      return index;
   }

   private byte[] readContent(int root, DataInputStream dataInput) throws IOException {
      // read length of decompressed text
      int length = dataInput.readInt();
      int readCharacters = 0;
      byte[] plaintext = new byte[length];
      XtendedDataInputStream bitIn = new XtendedDataInputStream(dataInput);

      do {
         int i = root;
         while (nodes[i].getLeft() != -1) {
            if (!bitIn.readBit()) {
               i = nodes[i].getLeft();
            }
            else {
               i = nodes[i].getRight();
            }
         }
         plaintext[readCharacters] = (byte)nodes[i].getValue();
         readCharacters++;

      } while (readCharacters < length);
      return plaintext;
   }


   private static class Node {

      private float frequency;
      private int parent;
      private int left;
      private int right;
      private int value;

      public Node(DataInput in) throws IOException {
         read(in);
      }

      public Node(int value) {
         this.value = value;
         this.frequency = 1.0f;
         this.left = -1;
         this.right = -1;
         this.parent = -1;
      }

      public Node(int left, int right, float frequency, int value) {
         this.left = left;
         this.right = right;
         this.frequency = frequency;
         this.parent = -2;
         this.value = value;
      }

      public float getFrequency() {
         return frequency;
      }

      public void setFrequency(float freq) {
         this.frequency = freq;
      }

      public int getParent() {
         return parent;
      }

      public void setParent(int parent) {
         this.parent = parent;
      }

      public int getLeft() {
         return left;
      }

      public void setLeft(int left) {
         this.left = left;
      }

      public int getRight() {
         return right;
      }

      public void setRight(int right) {
         this.right = right;
      }

      public int getValue() {
         return value;
      }

      public void setValue(int value) {
         this.value = value;
      }

      public void incFrequency() {
         frequency++;
      }

      public void normalizeFrequency(float divisor) {
         if (frequency != 0.0) frequency /= divisor;
      }

      public void read(DataInput in) throws IOException {
         left = in.readShort();
         right = in.readShort();
         value = in.readShort();
      }

      public void write(DataOutput out) throws IOException {
         out.writeShort(left);
         out.writeShort(right);
         out.writeShort(value);
      }
   }

}
