/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.xml;

import com.tagtraum.perf.io.XtendedDataInputStream;
import org.xml.sax.*;
import org.xml.sax.helpers.AttributesImpl;
import org.xml.sax.helpers.DefaultHandler;

import java.io.*;
import java.util.ArrayList;
import java.util.List;
import java.util.ResourceBundle;

/**
 * Decodes a stream of binary SAX events produced by {@link BSAXEncoder}.<p>
 * This class is not threadsafe.
 * <p>
 * The interface {@link XMLReader} is not completely implemented.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>
 * @version $Id: $
 * @see BSAXEncoder
 */
public class BSAXDecoder implements XMLReader {

   private ContentHandler contentHandler;
   private DTDHandler dtdHandler;
   private XtendedDataInputStream in;
   private String[][] symbolTable;
   private int currentCodePage;
   private int maxCode;
   private int maxCodePage;
   private List elementStack;
   private List namespaceStack;
   private static ResourceBundle localStrings = ResourceBundle.getBundle("com.tagtraum.perf.xml.localStrings");

   public BSAXDecoder() {
      elementStack = new ArrayList();
      namespaceStack = new ArrayList();
   }

   public void setDTDHandler(DTDHandler handler) {
      this.dtdHandler = handler;
   }

   public DTDHandler getDTDHandler() {
      return dtdHandler;
   }

   public void setContentHandler(ContentHandler handler) {
      contentHandler = handler;
   }

   public ContentHandler getContentHandler() {
      return contentHandler;
   }

   public void parse(String systemId)
         throws IOException, SAXException {
      throw new SAXException(localStrings.getString("inputsource_not_supported"));
   }


   public void parse(InputSource input) throws IOException, SAXException {
      InputStream inputStream = input.getByteStream();
      if (inputStream == null) throw new SAXException(localStrings.getString("inputsource_not_supported"));
      BufferedInputStream bin = (BufferedInputStream) (inputStream instanceof BufferedInputStream ? inputStream : new BufferedInputStream(inputStream));
      this.in = new XtendedDataInputStream(bin);
      symbolTable = null;
      elementStack.clear();
      namespaceStack.clear();
      maxCode = BSAXEncoder.RESERVED_CODES;

      int majorVersion = in.read();
      int minorVersion = in.read();
      if (majorVersion != 1 || minorVersion != 0) throw new SAXException(localStrings.getString("wrong_version"));

      // main loop
      int code;
      loop:
      while ((code = in.readHalfByte()) != -1) {
         //System.out.println("Decoder Code: " + code);
         switch (code) {
            case BSAXEncoder.NEW_BLOCK:
               readSymbolTable();
               break;
            case BSAXEncoder.START_DOCUMENT:
               if (contentHandler != null) contentHandler.startDocument();
               break;
            case BSAXEncoder.END_DOCUMENT:
               if (contentHandler != null) contentHandler.endDocument();
               break loop;
            case BSAXEncoder.START_ELEMENT:
               startElement();
               break;
            case BSAXEncoder.END_ELEMENT:
               endElement();
               break;
            case BSAXEncoder.CHARACTERS:
               characters();
               break;
            case BSAXEncoder.IGNORABLE_WHITESPACE:
               ignorableWhitespace();
               break;
            case BSAXEncoder.NOTATION_DECLARATION:
               if (dtdHandler != null) {
                  dtdHandler.notationDecl(in.readUTF(), in.readUTF(), in.readUTF());
               }
               else {
                  for (int i=0; i<3; i++) in.readUTF();
               }
               break;
            case BSAXEncoder.PROCESSING_INSTRUCTION:
               if (contentHandler != null) {
                  contentHandler.processingInstruction(in.readUTF(), in.readUTF());
               }
               else {
                  for (int i=0; i<2; i++) in.readUTF();
               }
               break;
            case BSAXEncoder.SKIPPED_ENTITY:
               if (contentHandler != null) {
                  contentHandler.skippedEntity(getSymbol(in.readHalfByte()));
               }
               else {
                  in.readHalfByte();
               }
               break;
            case BSAXEncoder.UNPARSED_ENTITY_DECLARATION:
               if (dtdHandler != null) {
                  dtdHandler.unparsedEntityDecl(in.readUTF(), in.readUTF(), in.readUTF(), in.readUTF());
               }
               else {
                  for (int i=0; i<4; i++) in.readUTF();
               }
               break;
            case BSAXEncoder.START_PREFIXMAPPING:
               if (contentHandler != null) {
                  PrefixMapping mapping = new PrefixMapping(getSymbol(in.readHalfByte()), getSymbol(in.readHalfByte()));
                  namespaceStack.add(mapping);
                  contentHandler.startPrefixMapping(mapping.getPrefix(), mapping.getURI());
               }
               else {
                  for (int i=0; i<2; i++) in.readHalfByte();
               }
               break;
            case BSAXEncoder.END_PREFIXMAPPING:
               if (contentHandler != null) {
                  String prefix = getSymbol(in.readHalfByte());
                  for (int i = namespaceStack.size() - 1; i >= 0; i--) {
                     if (((PrefixMapping) namespaceStack.get(i)).getPrefix().equals(prefix)) {
                        namespaceStack.remove(i);
                        break;
                     }
                  }
                  contentHandler.endPrefixMapping(prefix);
               }
               else {
                  in.readHalfByte();
               }
               break;
         }
      }
   }

   private void ignorableWhitespace() throws IOException, SAXException {
      if (contentHandler != null) {
         char[] whitespace = in.readUTF().toCharArray();
         contentHandler.ignorableWhitespace(whitespace, 0, whitespace.length);
      }
   }

   private void characters() throws IOException, SAXException {
      if (contentHandler != null) {
         char[] characters = in.readUTF().toCharArray();
         contentHandler.characters(characters, 0, characters.length);
      }
      return;
   }

   private void endElement() throws SAXException {
      if (contentHandler != null) {
         Element endElement = (Element) elementStack.get(elementStack.size() - 1);
         endElement.remove();
         contentHandler.endElement(endElement.getURI(), endElement.getLocalName(), endElement.getQName());
      }
   }

   private void startElement() throws IOException, SAXException {
      if (contentHandler != null) {
         String elementQName = getSymbol(in.readHalfByte());
         Attributes attributes = readAttributes();
         Element startElement = new Element(elementQName);
         contentHandler.startElement(startElement.getURI(), startElement.getLocalName(), startElement.getQName(), attributes);
      }
   }

   private Attributes readAttributes() throws IOException {
      int code;
      Attributes attributes = null;
      while ((code = in.readHalfByte()) != BSAXEncoder.END) {
         String qName = getSymbol(code);
         String localName;
         String uri;
         if (isNamespaceAware()) {
            int colon = qName.indexOf(':');
            if (colon == -1) {
               localName = qName;
               uri = getURI("");
            } else {
               localName = qName.substring(colon + 1);
               uri = getURI(qName.substring(0, colon));
            }
         } else {
            localName = "";
            uri = "";
         }
         if (attributes == null) attributes = new AttributesImpl();
         ((AttributesImpl) attributes).addAttribute(uri, localName, qName, getSymbol(in.readHalfByte()), in.readUTF());
      }
      if (attributes == null) attributes = EMPTY_ATTRIBUTES;
      return attributes;
   }

   private void readSymbolTable() throws IOException {
      int length = in.readInt();
      if (length > 0) {
         extendSymbolTable(length);
      }
      for (int i = 0; i < length; i++) {
         symbolTable[maxCodePage][maxCode - 2] = in.readUTF();
         //symbolTable.put(new SymbolKey(maxCodePage, maxCode), string);
         if (maxCode == 0x0f) {
            maxCode = 1;
            maxCodePage++;
         }
         maxCode++;
      }
   }

   private void extendSymbolTable(int toAdd) {
      if (symbolTable != null) {
         int currentLength = (maxCode - BSAXEncoder.RESERVED_CODES) + maxCodePage * BSAXEncoder.DEFINABLE_CODES_PER_CODEPAGE;
         String[][] newSymbolTable = new String[(int) Math.ceil(currentLength + toAdd / (double) (BSAXEncoder.DEFINABLE_CODES_PER_CODEPAGE))][];
         // move over the old arrays and create some new ones
         for (int i = 0; i < symbolTable.length; i++) {
            newSymbolTable[i] = symbolTable[i];
         }
         for (int i = symbolTable.length; i < newSymbolTable.length; i++) {
            newSymbolTable[i] = new String[BSAXEncoder.DEFINABLE_CODES_PER_CODEPAGE];
         }
         symbolTable = newSymbolTable;
      } else {
         symbolTable = new String[(int) Math.ceil(toAdd / (double) (BSAXEncoder.DEFINABLE_CODES_PER_CODEPAGE))][BSAXEncoder.DEFINABLE_CODES_PER_CODEPAGE];
      }
   }

   private String getSymbol(int code) throws IOException {
      if (code == BSAXEncoder.SWITCH_CODE_PAGE) {
         currentCodePage = in.readHalfByte();
         currentCodePage |= in.readHalfByte() << 4;
         code = in.readHalfByte();
      }
      //return (String) symbolTable.get(symbolSearchKey.set(currentCodePage, code));
      //System.out.println("code: " + (code-2) + ", codePage: " + currentCodePage);
      return symbolTable[currentCodePage][code - 2];
   }

   private boolean isNamespaceAware() {
      return !namespaceStack.isEmpty();
   }

   /**
    * Returns the URI for a namespace-prefix that has been declared in any element.
    */
   private String getURI(String prefix) {
      for (int i = namespaceStack.size() - 1; i >= 0; i--) {
         PrefixMapping mapping = (PrefixMapping) namespaceStack.get(i);
         if (mapping.getPrefix().equals(prefix)) return mapping.getURI();
      }
      return "";
   }

   public static void main(String[] args) throws Exception {
      BSAXDecoder decoder = new BSAXDecoder();
      DefaultHandler handler = new XMLWriter(new FileOutputStream(args[0] + ".decoded"), "ISO-8859-1", false);
      decoder.setDTDHandler(handler);
      decoder.setContentHandler(handler);
      decoder.parse(new InputSource(new FileInputStream(args[0])));
   }

   // Methods required by XMLReader interface

   public boolean getFeature(String name)
         throws SAXNotRecognizedException, SAXNotSupportedException {
      return false;
   }

   public void setFeature(String name, boolean value)
         throws SAXNotRecognizedException, SAXNotSupportedException {
   }

   public Object getProperty(String name)
         throws SAXNotRecognizedException, SAXNotSupportedException {
      return null;
   }

   public void setProperty(String name, Object value)
         throws SAXNotRecognizedException, SAXNotSupportedException {
   }

   public void setEntityResolver(EntityResolver resolver) {
   }

   public EntityResolver getEntityResolver() {
      return null;
   }

   public void setErrorHandler(ErrorHandler handler) {
   }

   public ErrorHandler getErrorHandler() {
      return null;
   }

   private class Element {

      private String localName;
      private String uri;
      private String qName;

      public Element(String qName) {
         this.qName = qName;
         if (isNamespaceAware()) {
            int colon = qName.indexOf(':');
            if (colon != -1) {
               this.localName = qName.substring(colon + 1);
               this.uri = BSAXDecoder.this.getURI(qName.substring(0, colon));
            } else {
               this.localName = qName;
               this.uri = BSAXDecoder.this.getURI("");
            }
         } else {
            this.localName = "";
            this.uri = "";
         }
         elementStack.add(this);
      }

      public void remove() {
         elementStack.remove(elementStack.size() - 1);
      }

      public String getLocalName() {
         return localName;
      }

      public String getURI() {
         return uri;
      }

      public String getQName() {
         return qName;
      }
   }

   private static class PrefixMapping {
      private String prefix;
      private String uri;

      public PrefixMapping(String prefix, String uri) {
         this.prefix = prefix;
         this.uri = uri;
      }

      public String getPrefix() {
         return prefix;
      }

      public String getURI() {
         return uri;
      }
   }

   private static final Attributes EMPTY_ATTRIBUTES = new Attributes() {
      public int getLength() {
         return 0;
      }

      public String getURI(int index) {
         return null;
      }

      public String getLocalName(int index) {
         return null;
      }

      public String getQName(int index) {
         return null;
      }

      public String getType(int index) {
         return null;
      }

      public String getValue(int index) {
         return null;
      }

      public int getIndex(String uri, String localPart) {
         return -1;
      }

      public int getIndex(String qName) {
         return -1;
      }

      public String getType(String uri, String localName) {
         return null;
      }

      public String getType(String qName) {
         return null;
      }

      public String getValue(String uri, String localName) {
         return null;
      }

      public String getValue(String qName) {
         return null;
      }
   };
}
