/*

Copyright (c) 2002 Hendrik Schreiber <hs@tagtraum.com>.

LGPL
====

This is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For LGPL see <http://www.fsf.org/copyleft/lesser.txt>

For more information about this software see
<http://www.tagtraum.com/performance/>

*/
package com.tagtraum.perf.xml;

import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.ext.LexicalHandler;
import org.xml.sax.helpers.DefaultHandler;

import java.io.*;
import java.util.ArrayList;
import java.util.List;

/**
 * Can be used to write XML documents. This class checks whether start and end tags are balanced and
 * will throw {@link SAXException}s, if this is not the case.
 *
 * @author <a href="mailto:hs@tagtraum.com">Hendrik Schreiber</a>, Andy Clark, IBM - who wrote the Xerces SAX Writer example this class is derived from
 * @version $Id: $
 */
public class XMLWriter extends DefaultHandler implements LexicalHandler {

   private PrintWriter out;
   private String encoding;
   private List stack;
   private boolean canonical;

   public XMLWriter(OutputStream out, String encoding, boolean canonical) throws UnsupportedEncodingException {
      this.encoding = encoding == null?"UTF-8":encoding;
      this.out = new PrintWriter(new BufferedWriter(new OutputStreamWriter(out, this.encoding)));
      this.canonical = canonical;
      stack = new ArrayList();
   }

   public void startDocument() throws SAXException {
      if (!canonical) {
         out.print("<?xml version=\"1.0\" encoding=\"");
         out.print(encoding);
         out.print("\"?>");
         out.println();
      }
   }

   /**
    * Returns the name of next element that should be closed.
    */
   public String getEndElement() {
      if (stack.isEmpty()) return null;
      return (String) stack.get(stack.size() - 1);
   }

   /**
    * Writes end tags for all opened elements.
    */
   private void endAllPendingElements() {
      while (!stack.isEmpty()) {
         out.print("</");
         out.print(stack.remove(stack.size() - 1));
         out.print('>');
      }
   }

   /**
    * Prints all pending end tags and flushes the stream.
    */
   public void endDocument() throws SAXException {
      endAllPendingElements();
      out.flush();
   }

   public void processingInstruction(String target, String data)
      throws SAXException {
      if (stack.size() > 0) {
         out.print("<?");
         out.print(target);
         if (data != null && data.length() > 0) {
            out.print(' ');
            out.print(data);
         }
         out.print("?>");
      }
   }

   public void startElement(String uri, String local, String raw, Attributes attrs) throws SAXException {
      stack.add(raw);
      out.print('<');
      out.print(raw);
      if (attrs != null) {
         int len = attrs.getLength();
         for (int i = 0; i < len; i++) {
            out.print(' ');
            out.print(attrs.getQName(i));
            out.print("=\"");
            normalizeAndPrint(attrs.getValue(i));
            out.print('"');
         }
      }
      out.print('>');

      out.flush();
   }

   /**
    * @exception SAXException if start and end tags aren't balanced properly
    */
   public void endElement(String uri, String local, String raw) throws SAXException {
      String expectedTag = (String) stack.remove(stack.size() - 1);
      if (!raw.equals(expectedTag)) throw new SAXException("Expected </" + expectedTag + "> instead of </" + raw + ">.");
      out.print("</");
      out.print(raw);
      out.print('>');

      out.flush();
   }

   public void characters(char ch[], int start, int length)
      throws SAXException {
      normalizeAndPrint(ch, start, length);
   }

   public void ignorableWhitespace(char ch[], int start, int length)
      throws SAXException {
      normalizeAndPrint(ch, start, length);
   }

   public void comment(char ch[], int start, int length)
      throws SAXException {
      if (!canonical && stack.size() > 0) {
          out.print("<!--");
          normalizeAndPrint(ch, start, length);
          out.print("-->");
      }
   }

   // necessary methods to please LexicalHandler

   public void startDTD(String name, String publicId,
                        String systemId)
      throws SAXException {
   }


   public void endDTD()
      throws SAXException {
   }
   public void startEntity(String name)
      throws SAXException {
   }

   public void endEntity(String name)
      throws SAXException {
   }

   public void startCDATA()
      throws SAXException {
   }

   public void endCDATA()
      throws SAXException {
   }

   // end necessary methods to please LexicalHandler

   /**
    * Flushes the outputstream.
    */
   public void flush() {
      out.flush();
   }

   private void normalizeAndPrint(String s) {
      for (int i = 0, len = (s != null) ? s.length() : 0; i < len; i++) {
         char c = s.charAt(i);
         normalizeAndPrint(c);
      }
   }

   private void normalizeAndPrint(char[] ch, int offset, int length) {
      for (int i = 0; i < length; i++) {
         normalizeAndPrint(ch[offset + i]);
      }
   }

   private void normalizeAndPrint(char c) {
      switch (c) {
         case '<':
            {
               out.print("&lt;");
               break;
            }
         case '>':
            {
               out.print("&gt;");
               break;
            }
         case '&':
            {
               out.print("&amp;");
               break;
            }
         case '"':
            {
               out.print("&quot;");
               break;
            }
         case '\r':
         case '\n':
            {
               if (canonical) {
                  out.print("&#");
                  out.print(Integer.toString(c));
                  out.print(';');
                  break;
               }
            }
         default:
            {
               out.print(c);
            }
      }
   }
}
